#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for the convergence testing module

Author: Carl Laird
"""
import io
import json
import pytest
import os
import os.path
from collections import OrderedDict

import pyomo.environ as pe
from pyomo.common.fileutils import this_file_dir
from pyomo.common.unittest import assertStructuredAlmostEqual
import idaes.core.util.convergence.convergence_base as cb
import idaes

# See if ipopt is available and set up solver
ipopt_available = pe.SolverFactory("ipopt").available()
ceval_fixedvar_mutableparam_str = (
    "idaes.core.util.convergence.tests."
    "conv_eval_classes.ConvEvalFixedVarMutableParam"
)
ceval_fixedvar_immutableparam_str = (
    "idaes.core.util.convergence.tests."
    "conv_eval_classes.ConvEvalFixedVarImmutableParam"
)
ceval_unfixedvar_mutableparam_str = (
    "idaes.core.util.convergence.tests."
    "conv_eval_classes.ConvEvalUnfixedVarMutableParam"
)

currdir = this_file_dir()
wrtdir = idaes.testing_directory


@pytest.mark.unit
def test_convergence_evaluation_specification_file_fixedvar_mutableparam():
    ceval_class = cb._class_import(ceval_fixedvar_mutableparam_str)
    ceval = ceval_class()
    import idaes.core.util.convergence.tests.conv_eval_classes as cev

    assert ceval.__class__ == cev.ConvEvalFixedVarMutableParam().__class__

    spec = ceval.get_specification()
    fname = os.path.join(wrtdir, "ceval_fixedvar_mutableparam.3.42.json")
    cb.write_sample_file(
        spec, fname, ceval_fixedvar_mutableparam_str, n_points=3, seed=42
    )

    baseline_fname = os.path.join(
        currdir, "ceval_fixedvar_mutableparam.3.42.baseline.json"
    )

    with open(fname) as FILE:
        test = json.load(FILE)
    with open(baseline_fname) as FILE:
        baseline = json.load(FILE)
    assertStructuredAlmostEqual(baseline, test, abstol=1e-8)

    if os.path.exists(fname):
        os.remove(fname)


@pytest.mark.unit
def test_convergence_evaluation_specification_file_unfixedvar_mutableparam():
    ceval_class = cb._class_import(ceval_unfixedvar_mutableparam_str)
    ceval = ceval_class()
    import idaes.core.util.convergence.tests.conv_eval_classes as cev

    assert ceval.__class__ == cev.ConvEvalUnfixedVarMutableParam().__class__

    spec = ceval.get_specification()
    fname = os.path.join(wrtdir, "ceval_unfixedvar_mutableparam.3.42.json")
    cb.write_sample_file(
        spec, fname, ceval_unfixedvar_mutableparam_str, n_points=3, seed=42
    )

    baseline_fname = os.path.join(
        currdir, "ceval_unfixedvar_mutableparam.3.42.baseline.json"
    )

    with open(fname) as FILE:
        test = json.load(FILE)
    with open(baseline_fname) as FILE:
        baseline = json.load(FILE)
    assertStructuredAlmostEqual(baseline, test, abstol=1e-8)

    # expect an exception because var is not fixed
    with pytest.raises(ValueError):
        (
            inputs,
            samples,
            global_results,
        ) = cb.run_convergence_evaluation_from_sample_file(fname)

    if os.path.exists(fname):
        os.remove(fname)


@pytest.mark.unit
def test_convergence_evaluation_stats_from_to():
    ceval_class = cb._class_import(ceval_fixedvar_mutableparam_str)
    ceval = ceval_class()
    import idaes.core.util.convergence.tests.conv_eval_classes as cev

    assert ceval.__class__ == cev.ConvEvalFixedVarMutableParam().__class__

    spec = ceval.get_specification()
    fname = os.path.join(wrtdir, "ceval_fixedvar_mutableparam.3.43.json")
    cb.write_sample_file(
        spec, fname, ceval_fixedvar_mutableparam_str, n_points=3, seed=43
    )

    inputs, samples, results = cb.run_convergence_evaluation_from_sample_file(fname)

    s = cb.save_convergence_statistics(inputs, results)
    d = s.to_dict()
    s2 = cb.Stats(from_dict=d)
    fname1 = os.path.join(wrtdir, "stats1.json")

    with open(fname1, "w") as f:
        s.to_json(f)
    with open(fname1) as FILE:
        baseline = json.load(FILE)

    buf = io.StringIO()
    s2.to_json(buf)
    test = json.loads(buf.getvalue())
    assertStructuredAlmostEqual(baseline, test, abstol=1e-8)

    s3 = cb.Stats(from_json=fname1)
    buf = io.StringIO()
    s3.to_json(buf)
    test = json.loads(buf.getvalue())
    assertStructuredAlmostEqual(baseline, test, abstol=1e-8)

    if os.path.exists(fname):
        os.remove(fname)
    if os.path.exists(fname1):
        os.remove(fname1)


@pytest.mark.skipif(not ipopt_available, reason="Ipopt solver not available")
@pytest.mark.unit
def test_convergence_evaluation_single():
    ceval_class = cb._class_import(ceval_fixedvar_mutableparam_str)
    ceval = ceval_class()
    import idaes.core.util.convergence.tests.conv_eval_classes as cev

    assert ceval.__class__ == cev.ConvEvalFixedVarMutableParam().__class__

    spec = ceval.get_specification()
    fname = os.path.join(wrtdir, "ceval_fixedvar_mutableparam.3.43.json")
    cb.write_sample_file(
        spec, fname, ceval_fixedvar_mutableparam_str, n_points=3, seed=43
    )
    s, c, i, rest, reg, i = cb.run_single_sample_from_sample_file(
        fname, name="Sample-1"
    )
    assert c == True


@pytest.mark.unit
def test_convergence_evaluation_specification_file_fixedvar_immutableparam():
    ceval_class = cb._class_import(ceval_fixedvar_immutableparam_str)
    ceval = ceval_class()
    import idaes.core.util.convergence.tests.conv_eval_classes as cev

    assert ceval.__class__ == cev.ConvEvalFixedVarImmutableParam().__class__

    spec = ceval.get_specification()
    fname = os.path.join(wrtdir, "ceval_fixedvar_immutableparam.3.42.json")
    cb.write_sample_file(
        spec, fname, ceval_fixedvar_immutableparam_str, n_points=3, seed=42
    )

    baseline_fname = os.path.join(
        currdir, "ceval_fixedvar_immutableparam.3.42.baseline.json"
    )

    with open(fname) as FILE:
        test = json.load(FILE)
    with open(baseline_fname) as FILE:
        baseline = json.load(FILE)
    assertStructuredAlmostEqual(baseline, test, abstol=1e-8)

    # expect an exception because param is not mutable
    with pytest.raises(ValueError):
        (
            inputs,
            samples,
            global_results,
        ) = cb.run_convergence_evaluation_from_sample_file(fname)

    if os.path.exists(fname):
        os.remove(fname)


@pytest.mark.skipif(not ipopt_available, reason="Ipopt solver not available")
@pytest.mark.unit
def test_convergence_evaluation_fixedvar_mutableparam():
    ceval_class = cb._class_import(ceval_fixedvar_mutableparam_str)
    ceval = ceval_class()
    import idaes.core.util.convergence.tests.conv_eval_classes as cev

    assert ceval.__class__ == cev.ConvEvalFixedVarMutableParam().__class__

    spec = ceval.get_specification()
    fname = os.path.join(currdir, "ceval_fixedvar_mutableparam.3.43.json")
    cb.write_sample_file(
        spec, fname, ceval_fixedvar_mutableparam_str, n_points=3, seed=43
    )

    inputs, samples, global_results = cb.run_convergence_evaluation_from_sample_file(
        fname
    )

    # put the results into a json file for comparison
    # jsondict = dict(inputs=inputs, samples=samples,
    #                 global_results=global_results)
    # results_fname = os.path.join(
    #        currdir, 'ceval_fixedvar_mutableparam.3.43.results.json')
    # with open(results_fname, 'w') as fd:
    #     json.dump(jsondict, fd, indent=3)

    # compare results
    assert global_results[0]["name"] == "Sample-1"
    assert global_results[0]["solved"]
    # This should take 14 iterations to converge with the IDAES solver
    # distribution, but due to various solver factors the number of iterations
    # could vary.
    assert global_results[0]["iters"] == pytest.approx(14, abs=2)

    assert global_results[1]["name"] == "Sample-2"
    assert global_results[1]["solved"]
    # This should take 15 iterations to converge with the IDAES solver
    # distribution, but due to various solver factors the number of iterations
    # could vary.
    assert global_results[1]["iters"] == pytest.approx(15, abs=2)

    assert global_results[2]["name"] == "Sample-3"
    assert global_results[2]["solved"]
    # This should take 12 iterations to converge with the IDAES solver
    # distribution, but due to various solver factors the number of iterations
    # could vary.
    assert global_results[2]["iters"] == pytest.approx(12, abs=2)

    if os.path.exists(fname):
        os.remove(fname)
    # if os.path.exists(results_fname):
    #     os.remove(results_fname)


@pytest.mark.unit
def test_parse_ipopt_output():
    fname = os.path.join(currdir, "ipopt_output.txt")
    iters, restoration, regularization, time = cb._parse_ipopt_output(fname)

    assert iters == 43
    assert restoration == 39
    assert regularization == 4
    assert time == 0.016 + 0.035


@pytest.mark.unit
def test_compare_convergence_runs_all_same():
    run1 = [
        OrderedDict(
            [
                ("name", "Sample_1"),
                ("solved", True),
                ("iters", 11),
                ("iters_in_restoration", 3),
                ("iters_w_regularization", 7),
                ("time", 21),
            ]
        ),
        OrderedDict(
            [
                ("name", "Sample_2"),
                ("solved", False),
                ("iters", 12),
                ("iters_in_restoration", 4),
                ("iters_w_regularization", 8),
                ("time", 22),
            ]
        ),
        OrderedDict(
            [
                ("name", "Sample_3"),
                ("solved", True),
                ("iters", 13),
                ("iters_in_restoration", 5),
                ("iters_w_regularization", 9),
                ("time", 23),
            ]
        ),
    ]

    baseline = OrderedDict(
        [
            (
                "Sample_1",
                OrderedDict(
                    [
                        ("solved", True),
                        ("iters", 11),
                        ("iters_in_restoration", 3),
                        ("iters_w_regularization", 7),
                        ("time", 21),
                    ]
                ),
            ),
            (
                "Sample_2",
                OrderedDict(
                    [
                        ("solved", False),
                        ("iters", 12),
                        ("iters_in_restoration", 4),
                        ("iters_w_regularization", 8),
                        ("time", 22),
                    ]
                ),
            ),
            (
                "Sample_3",
                OrderedDict(
                    [
                        ("solved", True),
                        ("iters", 13),
                        ("iters_in_restoration", 5),
                        ("iters_w_regularization", 9),
                        ("time", 23),
                    ]
                ),
            ),
        ]
    )

    diff_solves, diff_iters, diff_rest, diff_reg = cb._compare_run_to_baseline(
        run1, baseline
    )

    assert diff_solves == []
    assert diff_iters == []
    assert diff_rest == []
    assert diff_reg == []


@pytest.mark.unit
def test_compare_convergence_runs_invalid_sample_name():
    run1 = [
        OrderedDict(
            [
                ("name", "Sample_X"),
                ("solved", True),
                ("iters", 11),
                ("iters_in_restoration", 3),
                ("iters_w_regularization", 7),
                ("time", 21),
            ]
        ),
        OrderedDict(
            [
                ("name", "Sample_2"),
                ("solved", False),
                ("iters", 12),
                ("iters_in_restoration", 4),
                ("iters_w_regularization", 8),
                ("time", 22),
            ]
        ),
        OrderedDict(
            [
                ("name", "Sample_3"),
                ("solved", True),
                ("iters", 13),
                ("iters_in_restoration", 5),
                ("iters_w_regularization", 9),
                ("time", 23),
            ]
        ),
    ]

    baseline = OrderedDict(
        [
            (
                "Sample_1",
                OrderedDict(
                    [
                        ("solved", True),
                        ("iters", 11),
                        ("iters_in_restoration", 3),
                        ("iters_w_regularization", 7),
                        ("time", 21),
                    ]
                ),
            ),
            (
                "Sample_2",
                OrderedDict(
                    [
                        ("solved", False),
                        ("iters", 12),
                        ("iters_in_restoration", 4),
                        ("iters_w_regularization", 8),
                        ("time", 22),
                    ]
                ),
            ),
            (
                "Sample_3",
                OrderedDict(
                    [
                        ("solved", True),
                        ("iters", 13),
                        ("iters_in_restoration", 5),
                        ("iters_w_regularization", 9),
                        ("time", 23),
                    ]
                ),
            ),
        ]
    )

    with pytest.raises(
        KeyError,
        match="baseline does not contain a sample with the name Sample_X. Please check that "
        "both convergence evaluation runs used the same set of samples.",
    ):
        diff_solves, diff_iters, diff_rest, diff_reg = cb._compare_run_to_baseline(
            run1, baseline
        )


@pytest.mark.unit
def test_compare_convergence_runs_differences_abs():
    run1 = [
        OrderedDict(
            [
                ("name", "Sample_1"),
                ("solved", True),  # diff
                ("iters", 11),  # same
                ("iters_in_restoration", 5),  # diff
                ("iters_w_regularization", 8),  # in tol
                ("time", 21),
            ]
        ),
        OrderedDict(
            [
                ("name", "Sample_2"),
                ("solved", False),  # diff
                ("iters", 13),  # in tol
                ("iters_in_restoration", 3),  # in tol
                ("iters_w_regularization", 6),  # diff
                ("time", 22),
            ]
        ),
        OrderedDict(
            [
                ("name", "Sample_3"),
                ("solved", True),  # same
                ("iters", 15),  # diff
                ("iters_in_restoration", 5),  # same
                ("iters_w_regularization", 9),  # same
                ("time", 23),
            ]
        ),
    ]

    baseline = OrderedDict(
        [
            (
                "Sample_1",
                OrderedDict(
                    [
                        ("solved", False),
                        ("iters", 11),
                        ("iters_in_restoration", 3),
                        ("iters_w_regularization", 7),
                        ("time", 21),
                    ]
                ),
            ),
            (
                "Sample_2",
                OrderedDict(
                    [
                        ("solved", True),
                        ("iters", 12),
                        ("iters_in_restoration", 4),
                        ("iters_w_regularization", 8),
                        ("time", 22),
                    ]
                ),
            ),
            (
                "Sample_3",
                OrderedDict(
                    [
                        ("solved", True),
                        ("iters", 13),
                        ("iters_in_restoration", 5),
                        ("iters_w_regularization", 9),
                        ("time", 23),
                    ]
                ),
            ),
        ]
    )

    diff_solves, diff_iters, diff_rest, diff_reg = cb._compare_run_to_baseline(
        run1, baseline, rel_tol=1e-5, abs_tol=1
    )

    assert diff_solves == ["Sample_1", "Sample_2"]
    assert diff_iters == ["Sample_3"]
    assert diff_rest == ["Sample_1"]
    assert diff_reg == ["Sample_2"]


@pytest.mark.unit
def test_compare_convergence_runs_differences_rel():
    run1 = [
        OrderedDict(
            [
                ("name", "Sample_1"),
                ("solved", True),  # diff
                ("iters", 11),  # same
                ("iters_in_restoration", 5),  # diff
                ("iters_w_regularization", 8),  # in tol
                ("time", 21),
            ]
        ),
        OrderedDict(
            [
                ("name", "Sample_2"),
                ("solved", True),  # diff
                ("iters", 13),  # in tol
                ("iters_in_restoration", 4),  # same
                ("iters_w_regularization", 4),  # diff
                ("time", 22),
            ]
        ),
        OrderedDict(
            [
                ("name", "Sample_3"),
                ("solved", False),  # diff
                ("iters", 20),  # diff
                ("iters_in_restoration", 4),  # in tol
                ("iters_w_regularization", 9),  # same
                ("time", 23),
            ]
        ),
    ]

    baseline = OrderedDict(
        [
            (
                "Sample_1",
                OrderedDict(
                    [
                        ("solved", False),
                        ("iters", 11),
                        ("iters_in_restoration", 3),
                        ("iters_w_regularization", 7),
                        ("time", 21),
                    ]
                ),
            ),
            (
                "Sample_2",
                OrderedDict(
                    [
                        ("solved", True),
                        ("iters", 12),
                        ("iters_in_restoration", 4),
                        ("iters_w_regularization", 8),
                        ("time", 22),
                    ]
                ),
            ),
            (
                "Sample_3",
                OrderedDict(
                    [
                        ("solved", True),
                        ("iters", 13),
                        ("iters_in_restoration", 5),
                        ("iters_w_regularization", 9),
                        ("time", 23),
                    ]
                ),
            ),
        ]
    )

    diff_solves, diff_iters, diff_rest, diff_reg = cb._compare_run_to_baseline(
        run1, baseline, rel_tol=0.25, abs_tol=0
    )

    assert diff_solves == ["Sample_1", "Sample_3"]
    assert diff_iters == ["Sample_3"]
    assert diff_rest == ["Sample_1"]
    assert diff_reg == ["Sample_2"]



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

import pytest
from pyomo.common import unittest as pyo_unittest
from idaes.apps.grid_integration.forecaster import ForecastError, Backcaster
import idaes.logger as idaeslog


@pytest.fixture
def historical_da_prices():
    return {"test_bus": [1] * 24 + [2] * 24 + [3] * 24}


@pytest.fixture
def historical_rt_prices():
    return {"test_bus": [10] * 24 + [20] * 24 + [30] * 24}


@pytest.fixture
def base_backcaster(historical_da_prices, historical_rt_prices):
    return Backcaster(historical_da_prices, historical_rt_prices)


@pytest.mark.unit
def test_create_backcaster(historical_da_prices, historical_rt_prices):
    backcaster = Backcaster(historical_da_prices, historical_rt_prices)
    assert backcaster.historical_da_prices is historical_da_prices
    assert backcaster.historical_rt_prices is historical_rt_prices


@pytest.mark.unit
def test_create_backcaster_with_small_max_historical_days(
    caplog, historical_da_prices, historical_rt_prices
):
    max_n_days = 1
    with caplog.at_level(idaeslog.WARNING):
        backcaster = Backcaster(
            historical_da_prices, historical_rt_prices, max_historical_days=max_n_days
        )

    _warn_msg = (
        f"The number of days in the input historical prices for bus test_bus is greater than the max value 1."
        f" Dropping the data for the first 2 day(s)."
    )
    assert _warn_msg in caplog.text

    expected_historical_da_prices = {"test_bus": [3] * 24}
    expected_historical_rt_prices = {"test_bus": [30] * 24}

    pyo_unittest.assertStructuredAlmostEqual(
        first=backcaster.historical_da_prices, second=expected_historical_da_prices
    )

    pyo_unittest.assertStructuredAlmostEqual(
        first=backcaster.historical_rt_prices, second=expected_historical_rt_prices
    )


@pytest.mark.unit
@pytest.mark.parametrize("value", ["10", "ten", [10]])
def test_invalid_max_historical_days_type(
    value, historical_da_prices, historical_rt_prices
):
    with pytest.raises(TypeError, match=r".*max_historical_days must be a number.*"):
        Backcaster(
            historical_da_prices, historical_rt_prices, max_historical_days=value
        )


@pytest.mark.unit
@pytest.mark.parametrize("value", [-10, 0, -0.0])
def test_invalid_max_historical_days_value(
    value, historical_da_prices, historical_rt_prices
):
    with pytest.raises(ValueError, match=r".*max_historical_days must be >= 1.*"):
        Backcaster(
            historical_da_prices, historical_rt_prices, max_historical_days=value
        )


@pytest.mark.unit
def test_create_backcaster_with_non_dict(historical_da_prices, historical_rt_prices):
    with pytest.raises(
        TypeError, match=r"Given historical price is not an dictionary object.*"
    ):
        Backcaster(historical_da_prices, [])
        Backcaster(100, historical_rt_prices)


@pytest.mark.unit
def test_create_backcaster_with_empty_dict(historical_da_prices, historical_rt_prices):
    with pytest.raises(ValueError, match=r"Given historical price is empty."):
        Backcaster({}, historical_rt_prices)
        Backcaster(historical_da_prices, {})


@pytest.mark.unit
def test_create_backcaster_with_non_list(historical_da_prices, historical_rt_prices):
    with pytest.raises(TypeError, match=r".*bus test_bus is not a list object.*"):
        Backcaster({"test_bus": {1, 2, 3}}, historical_rt_prices)
        Backcaster(historical_da_prices, {"test_bus": {1, 2, 3}})


@pytest.mark.unit
def test_create_backcaster_with_not_enough_entries(
    historical_da_prices, historical_rt_prices
):
    with pytest.raises(
        ValueError, match=r".*At least a day of the historical prices.*"
    ):
        Backcaster({"test_bus": [1] * 23}, historical_rt_prices)
        Backcaster(historical_da_prices, {"test_bus": [1] * 2})


@pytest.mark.unit
def test_create_backcaster_with_non_24_entries(
    historical_da_prices, historical_rt_prices
):
    with pytest.raises(ValueError, match=r".*should be a multiple of 24.*"):
        Backcaster({"test_bus": [1] * 47}, historical_rt_prices)
        Backcaster(historical_da_prices, {"test_bus": [1] * 46})


@pytest.mark.unit
def test_forecast_real_time_prices(base_backcaster):

    n_samples = 3
    horizon = 4

    result_forecasts = base_backcaster.forecast_real_time_prices(
        date="2022-05-11", hour=18, bus="test_bus", horizon=horizon, n_samples=n_samples
    )
    expected_forecasts = {
        n_samples - i - 1: [(i + 1) * 10] * horizon for i in range(n_samples)
    }

    pyo_unittest.assertStructuredAlmostEqual(
        first=result_forecasts, second=expected_forecasts
    )


@pytest.mark.unit
def test_forecast_day_ahead_prices(base_backcaster):

    n_samples = 2
    horizon = 48

    result_forecasts = base_backcaster.forecast_day_ahead_prices(
        date="2022-05-11", hour=0, bus="test_bus", horizon=horizon, n_samples=n_samples
    )
    expected_forecasts = {0: [3] * 24 + [1] * 24, 1: [2] * 24 + [3] * 24}

    pyo_unittest.assertStructuredAlmostEqual(
        first=result_forecasts, second=expected_forecasts
    )


@pytest.mark.unit
def test_forecast_day_ahead_and_real_time_prices(base_backcaster):

    (
        da_result_forecasts,
        rt_forecasts,
    ) = base_backcaster.forecast_day_ahead_and_real_time_prices(
        date="2022-05-11", hour=0, bus="test_bus", horizon=48, n_samples=2
    )
    expected_da_forecasts = {0: [3] * 24 + [1] * 24, 1: [2] * 24 + [3] * 24}

    pyo_unittest.assertStructuredAlmostEqual(
        first=da_result_forecasts, second=expected_da_forecasts
    )

    (
        da_forecasts,
        rt_result_forecasts,
    ) = base_backcaster.forecast_day_ahead_and_real_time_prices(
        date="2022-05-11", hour=18, bus="test_bus", horizon=4, n_samples=3
    )

    expected_rt_forecasts = {3 - i - 1: [(i + 1) * 10] * 4 for i in range(3)}
    pyo_unittest.assertStructuredAlmostEqual(
        first=rt_result_forecasts, second=expected_rt_forecasts
    )


@pytest.mark.unit
def test_forecast_nonexistent_bus_prices(base_backcaster):

    wrong_bus = "test_bussss"

    n_samples = 3
    horizon = 4

    with pytest.raises(
        ForecastError, match=r"No test_bussss real-time price available"
    ):
        base_backcaster.forecast_real_time_prices(
            date="2022-05-11",
            hour=18,
            bus=wrong_bus,
            horizon=horizon,
            n_samples=n_samples,
        )

    n_samples = 2
    horizon = 48

    with pytest.raises(
        ForecastError, match=r"No test_bussss day-ahead price available"
    ):
        result_forecasts = base_backcaster.forecast_day_ahead_prices(
            date="2022-05-11",
            hour=0,
            bus=wrong_bus,
            horizon=horizon,
            n_samples=n_samples,
        )


class MockPrescientHourlyStats:
    def __init__(self, bus_to_prices_dict) -> None:
        self.observed_bus_LMPs = bus_to_prices_dict


@pytest.mark.unit
def test_fetch_hourly_stats_from_prescient(base_backcaster, historical_rt_prices):

    prescient_hourly_stats = MockPrescientHourlyStats({"test_bus": 15})
    base_backcaster.fetch_hourly_stats_from_prescient(prescient_hourly_stats)

    expected_current_day_rt_prices = {}
    expected_current_day_rt_prices["test_bus"] = [15]

    pyo_unittest.assertStructuredAlmostEqual(
        first=expected_current_day_rt_prices["test_bus"],
        second=base_backcaster._current_day_rt_prices["test_bus"],
    )

    for i in range(23):
        prescient_hourly_stats.observed_bus_LMPs["test_bus"] = 15
        base_backcaster.fetch_hourly_stats_from_prescient(prescient_hourly_stats)

    expected_current_day_rt_prices_1 = {}
    expected_current_day_rt_prices_1["test_bus"] = []

    expected_historical_rt_prices = [10] * 24 + [20] * 24 + [30] * 24 + [15] * 24

    pyo_unittest.assertStructuredAlmostEqual(
        first=expected_current_day_rt_prices_1["test_bus"],
        second=base_backcaster._current_day_rt_prices["test_bus"],
    )
    pyo_unittest.assertStructuredAlmostEqual(
        first=expected_historical_rt_prices,
        second=base_backcaster._historical_rt_prices["test_bus"],
    )


@pytest.mark.unit
def test_fetch_hourly_stats_from_prescient_greater_than_max_historical_days(
    base_backcaster, historical_rt_prices
):

    days = 8
    target_lmp = []
    for day in range(days):
        for t in range(24):
            prescient_hourly_stats = MockPrescientHourlyStats({"test_bus": day * 10})
            base_backcaster.fetch_hourly_stats_from_prescient(prescient_hourly_stats)
            target_lmp.append(day * 10)

    expected_historical_rt_prices = [20] * 24 + [30] * 24 + target_lmp

    pyo_unittest.assertStructuredAlmostEqual(
        first=expected_historical_rt_prices,
        second=base_backcaster._historical_rt_prices["test_bus"],
    )


class DAPrices:
    def __init__(self, da_prices) -> None:
        self.day_ahead_prices = da_prices

    def get(self, info):
        bus, t = info
        return self.day_ahead_prices[bus][t]


class MockRucMarket:
    def __init__(self, da_prices) -> None:
        self.day_ahead_prices = DAPrices(da_prices)


class MockPrescientRucPlan:
    def __init__(self, da_prices) -> None:
        self.ruc_market = MockRucMarket(da_prices)


@pytest.mark.unit
def test_fetch_day_ahead_stats_from_prescient(base_backcaster, historical_da_prices):

    for i in range(base_backcaster.max_historical_days + 1):
        da_price = {"test_bus": [i] * 24}
        day_ahead_result = MockPrescientRucPlan(da_price)
        base_backcaster.fetch_day_ahead_stats_from_prescient(
            None, None, day_ahead_result
        )

    expected_historical_da_prices = {}
    expected_historical_da_prices["test_bus"] = []
    for i in range(1, base_backcaster.max_historical_days + 1):
        expected_historical_da_prices["test_bus"] += [i] * 24

    pyo_unittest.assertStructuredAlmostEqual(
        first=expected_historical_da_prices,
        second=base_backcaster._historical_da_prices,
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Reboiler unit model. Tests 2 sets of state vars using the
ideal property package (FTPz and FcTP).

Author: Jaffer Ghouse
"""
import pytest
from pyomo.environ import check_optimal_termination, ConcreteModel, value
from pyomo.util.check_units import assert_units_consistent

from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)
from idaes.models_extra.column_models import Reboiler
from idaes.models.properties.activity_coeff_models.BTX_activity_coeff_VLE import (
    BTXParameterBlock,
)
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.util.testing import PhysicalParameterTestBlock, initialization_tester
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


@pytest.mark.unit
def test_config():

    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = Reboiler(property_package=m.fs.properties)

    assert len(m.fs.unit.config) == 9

    assert not m.fs.unit.config.has_boilup_ratio
    assert m.fs.unit.config.material_balance_type == MaterialBalanceType.useDefault
    assert m.fs.unit.config.energy_balance_type == EnergyBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert hasattr(m.fs.unit, "heat_duty")


class TestBTXIdeal:
    @pytest.fixture(scope="class")
    def btx_ftpz(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = BTXParameterBlock(
            valid_phase=("Liq", "Vap"), activity_coeff_model="Ideal"
        )

        m.fs.unit = Reboiler(property_package=m.fs.properties, has_boilup_ratio=True)

        # Fix the reboiler variables
        m.fs.unit.boilup_ratio.fix(1)

        # Fix the inputs (typically this will be the outlet liquid from the
        # bottom tray)
        m.fs.unit.inlet.flow_mol.fix(1)
        m.fs.unit.inlet.temperature.fix(363)
        m.fs.unit.inlet.pressure.fix(101325)
        m.fs.unit.inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        m.fs.unit.inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        iscale.calculate_scaling_factors(m)

        return m

    @pytest.fixture(scope="class")
    def btx_fctp(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = BTXParameterBlock(
            valid_phase=("Liq", "Vap"), activity_coeff_model="Ideal", state_vars="FcTP"
        )

        m.fs.unit = Reboiler(property_package=m.fs.properties, has_boilup_ratio=True)

        # Fix the reboiler variables
        m.fs.unit.boilup_ratio.fix(1)

        # Fix the inputs (typically this will be the outlet liquid from the
        # bottom tray)
        m.fs.unit.inlet.flow_mol_comp[0, "benzene"].fix(0.5)
        m.fs.unit.inlet.flow_mol_comp[0, "toluene"].fix(0.5)
        m.fs.unit.inlet.temperature.fix(363)
        m.fs.unit.inlet.pressure.fix(101325)

        iscale.calculate_scaling_factors(m)

        return m

    @pytest.mark.unit
    def test_build(self, btx_ftpz, btx_fctp):

        assert hasattr(btx_ftpz.fs.unit, "boilup_ratio")
        assert hasattr(btx_ftpz.fs.unit, "eq_boilup_ratio")

        assert hasattr(btx_ftpz.fs.unit, "inlet")
        assert hasattr(btx_ftpz.fs.unit.inlet, "flow_mol")
        assert hasattr(btx_ftpz.fs.unit.inlet, "mole_frac_comp")
        assert hasattr(btx_ftpz.fs.unit.inlet, "temperature")
        assert hasattr(btx_ftpz.fs.unit.inlet, "pressure")

        assert hasattr(btx_ftpz.fs.unit, "bottoms")
        assert hasattr(btx_ftpz.fs.unit.bottoms, "flow_mol")
        assert hasattr(btx_ftpz.fs.unit.bottoms, "mole_frac_comp")
        assert hasattr(btx_ftpz.fs.unit.bottoms, "temperature")
        assert hasattr(btx_ftpz.fs.unit.bottoms, "pressure")

        assert hasattr(btx_ftpz.fs.unit, "vapor_reboil")
        assert hasattr(btx_ftpz.fs.unit.vapor_reboil, "flow_mol")
        assert hasattr(btx_ftpz.fs.unit.vapor_reboil, "mole_frac_comp")
        assert hasattr(btx_ftpz.fs.unit.vapor_reboil, "temperature")
        assert hasattr(btx_ftpz.fs.unit.vapor_reboil, "pressure")

        assert number_variables(btx_ftpz.fs.unit) == 48
        assert number_total_constraints(btx_ftpz.fs.unit) == 42
        assert number_unused_variables(btx_ftpz) == 0

        assert hasattr(btx_fctp.fs.unit, "boilup_ratio")
        assert hasattr(btx_fctp.fs.unit, "eq_boilup_ratio")

        assert hasattr(btx_fctp.fs.unit, "inlet")
        assert hasattr(btx_fctp.fs.unit.inlet, "flow_mol_comp")
        assert hasattr(btx_fctp.fs.unit.inlet, "temperature")
        assert hasattr(btx_fctp.fs.unit.inlet, "pressure")

        assert hasattr(btx_fctp.fs.unit, "bottoms")
        assert hasattr(btx_fctp.fs.unit.bottoms, "flow_mol_comp")
        assert hasattr(btx_fctp.fs.unit.bottoms, "temperature")
        assert hasattr(btx_fctp.fs.unit.bottoms, "pressure")

        assert hasattr(btx_fctp.fs.unit, "vapor_reboil")
        assert hasattr(btx_fctp.fs.unit.vapor_reboil, "flow_mol_comp")
        assert hasattr(btx_fctp.fs.unit.vapor_reboil, "temperature")
        assert hasattr(btx_fctp.fs.unit.vapor_reboil, "pressure")

        assert number_variables(btx_fctp.fs.unit) == 50
        assert number_total_constraints(btx_fctp.fs.unit) == 45
        assert number_unused_variables(btx_fctp) == 0

    @pytest.mark.unit
    def test_dof(self, btx_ftpz, btx_fctp):
        assert degrees_of_freedom(btx_ftpz) == 0
        assert degrees_of_freedom(btx_fctp) == 0

    @pytest.mark.component
    def test_units_FTPz(self, btx_ftpz, btx_fctp):
        assert_units_consistent(btx_ftpz)

    @pytest.mark.component
    def test_units_FcTP(self, btx_fctp):
        assert_units_consistent(btx_fctp)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, btx_ftpz, btx_fctp):
        initialization_tester(btx_ftpz)
        initialization_tester(btx_fctp)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, btx_ftpz, btx_fctp):
        results = solver.solve(btx_ftpz)

        # Check for optimal solution
        assert check_optimal_termination(results)

        results = solver.solve(btx_fctp)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, btx_ftpz, btx_fctp):

        # from idaes.core.util.scaling import unscaled_constraints_generator, unscaled_variables_generator, badly_scaled_var_generator
        # print("\nUnscaled Constraints")
        # for i in unscaled_constraints_generator(btx_ftpz):
        #     print(i)
        # print("\nUnscaled Vars")
        # for i in unscaled_variables_generator(btx_ftpz):
        #     print(i)
        # print("\nBadly Scaled Vars")
        # for i in badly_scaled_var_generator(btx_ftpz):
        #     print(i)
        # assert False

        # Bottoms port
        assert pytest.approx(0.5, abs=1e-3) == value(
            btx_ftpz.fs.unit.bottoms.flow_mol[0]
        )
        assert pytest.approx(0.3891, abs=1e-3) == value(
            btx_ftpz.fs.unit.bottoms.mole_frac_comp[0, "benzene"]
        )
        assert pytest.approx(0.6109, abs=1e-3) == value(
            btx_ftpz.fs.unit.bottoms.mole_frac_comp[0, "toluene"]
        )
        assert pytest.approx(368.728, abs=1e-2) == value(
            btx_ftpz.fs.unit.bottoms.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            btx_ftpz.fs.unit.bottoms.pressure[0]
        )

        # Vapor reboil port
        assert pytest.approx(0.5, abs=1e-3) == value(
            btx_ftpz.fs.unit.vapor_reboil.flow_mol[0]
        )
        assert pytest.approx(0.6108, abs=1e-3) == value(
            btx_ftpz.fs.unit.vapor_reboil.mole_frac_comp[0, "benzene"]
        )
        assert pytest.approx(0.3892, abs=1e-3) == value(
            btx_ftpz.fs.unit.vapor_reboil.mole_frac_comp[0, "toluene"]
        )
        assert pytest.approx(368.728, abs=1e-2) == value(
            btx_ftpz.fs.unit.vapor_reboil.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            btx_ftpz.fs.unit.vapor_reboil.pressure[0]
        )

        # Unit level
        assert pytest.approx(16926.5, rel=1e-4) == value(btx_ftpz.fs.unit.heat_duty[0])

        # Reboiler when using FcTP

        # Bottoms port
        assert pytest.approx(0.19455, abs=1e-3) == value(
            btx_fctp.fs.unit.bottoms.flow_mol_comp[0, "benzene"]
        )
        assert pytest.approx(0.30545, abs=1e-3) == value(
            btx_fctp.fs.unit.bottoms.flow_mol_comp[0, "toluene"]
        )
        assert pytest.approx(368.728, abs=1e-2) == value(
            btx_fctp.fs.unit.bottoms.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            btx_fctp.fs.unit.bottoms.pressure[0]
        )

        # Vapor reboil port
        assert pytest.approx(0.3054, abs=1e-3) == value(
            btx_fctp.fs.unit.vapor_reboil.flow_mol_comp[0, "benzene"]
        )
        assert pytest.approx(0.1946, abs=1e-3) == value(
            btx_fctp.fs.unit.vapor_reboil.flow_mol_comp[0, "toluene"]
        )
        assert pytest.approx(368.728, abs=1e-2) == value(
            btx_fctp.fs.unit.vapor_reboil.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            btx_fctp.fs.unit.vapor_reboil.pressure[0]
        )

        # Unit level
        assert pytest.approx(16926.5, rel=1e-4) == value(btx_fctp.fs.unit.heat_duty[0])

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, btx_ftpz, btx_fctp):
        assert (
            abs(
                value(
                    btx_ftpz.fs.unit.inlet.flow_mol[0]
                    - (
                        btx_ftpz.fs.unit.bottoms.flow_mol[0]
                        + btx_ftpz.fs.unit.vapor_reboil.flow_mol[0]
                    )
                )
            )
            <= 1e-6
        )

        assert (
            abs(
                value(
                    btx_fctp.fs.unit.inlet.flow_mol_comp[0, "benzene"]
                    + btx_fctp.fs.unit.inlet.flow_mol_comp[0, "toluene"]
                    - (
                        btx_fctp.fs.unit.bottoms.flow_mol_comp[0, "benzene"]
                        + btx_fctp.fs.unit.bottoms.flow_mol_comp[0, "toluene"]
                        + btx_fctp.fs.unit.vapor_reboil.flow_mol_comp[0, "benzene"]
                        + btx_fctp.fs.unit.vapor_reboil.flow_mol_comp[0, "toluene"]
                    )
                )
            )
            <= 1e-6
        )

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, btx_ftpz, btx_fctp):
        btx_ftpz.fs.unit.report()
        btx_fctp.fs.unit.report()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Drum model
The drum model consists of three main unit operations
1) a flash model (only for water and steam)
2) a mixer model
3) a 1-D water tank model with leevl

Inlet Ports:

* water/steam mixture from water wall
* feedwater_inlet: feedwater from economizer/pipe

Outlet Ports:

* liquid_outlet: liquid to downcomer
* steam_outlet: steam leaving the drum

The drum model receives saturated water from the water wall, and this is
separated at the flash unit, the steam leaves the drum, while the water/liquid
mixes with the feedwater from the economizer (or water pipe model).
Finally, the mixed state liquid (stream leaving the mixer), is used as the
control volume of the drum level model, which computes velocity and pressure
drop. The exit of the drum model is the liquid outlet.

Created on October 27 2020 by Boiler Team (J. Ma, M. Zamarripa)
"""
import pytest

# Import Pyomo libraries
import pyomo.environ as pyo

# Import IDAES core
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom

# Import Unit Model Modules
from idaes.models.properties import iapws95

# from idaes.models_extra.power_generation.unit_models.drum_1D import Drum1D
from idaes.models_extra.power_generation.unit_models.drum1D import Drum1D
import idaes.core.util.scaling as iscale
from idaes.core.util.testing import initialization_tester
from idaes.core.solvers import get_solver

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------


@pytest.fixture(scope="module")
def build_drum1D():
    # Create a Concrete Model as the top level object
    m = pyo.ConcreteModel()
    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(dynamic=False)
    # Add property packages to flowsheet library
    m.fs.prop_water = iapws95.Iapws95ParameterBlock()
    m.fs.unit = Drum1D(
        property_package=m.fs.prop_water,
        has_holdup=True,
        has_heat_transfer=True,
        has_pressure_change=True,
        finite_elements=4,
        drum_inner_diameter=1.2,
        drum_thickness=0.119,
    )

    m.fs.unit.drum_length.fix(15.3256)
    m.fs.unit.level[:].fix(0.6)
    m.fs.unit.number_downcomer.fix(6)
    m.fs.unit.downcomer_diameter.fix(0.38)
    m.fs.unit.temperature_ambient[:].fix(298.15)
    m.fs.unit.insulation_thickness.fix(0.145)

    return m


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.unit
def test_basic_build(build_drum1D):
    """Make a turbine model and make sure it doesn't throw exception"""
    m = build_drum1D
    assert degrees_of_freedom(m) == 5
    # Check unit config arguments
    assert len(m.fs.unit.config) == 13
    assert m.fs.unit.config.has_heat_transfer
    assert m.fs.unit.config.has_pressure_change
    assert m.fs.unit.config.property_package is m.fs.prop_water


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize_drum1D(build_drum1D):
    m = build_drum1D
    iscale.calculate_scaling_factors(m)
    state_args_water_steam = {
        "flow_mol": 14409.02,  # mol/s
        "pressure": 12024201.99,  # Pa
        "enth_mol": 28365.2608,
    }  # j/mol

    state_args_feedwater = {
        "flow_mol": 11554.58,
        "pressure": 12024201.99,
        "enth_mol": 22723.907,
    }

    initialization_tester(
        build_drum1D,
        dof=5,
        state_args_water_steam=state_args_water_steam,
        state_args_feedwater=state_args_feedwater,
    )


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_run_drum1D(build_drum1D):
    m = build_drum1D
    # fix inlets
    m.fs.unit.water_steam_inlet.flow_mol[:].fix()
    m.fs.unit.water_steam_inlet.pressure[:].fix()
    m.fs.unit.water_steam_inlet.enth_mol[:].fix()
    m.fs.unit.feedwater_inlet.flow_mol[:].fix()
    m.fs.unit.feedwater_inlet.enth_mol[:].fix()

    optarg = {"tol": 1e-7, "max_iter": 40}
    solver.options = optarg
    # solve model
    results = solver.solve(m, tee=True)
    # Check for optimal solution
    assert pyo.check_optimal_termination(results)
    assert degrees_of_freedom(m) == 0
    assert pytest.approx(0.6, abs=1e-3) == pyo.value(m.fs.unit.level[0])
    # energy balance
    assert pytest.approx(0, abs=1e-3) == pyo.value(
        m.fs.unit.water_steam_inlet.flow_mol[0]
        * m.fs.unit.water_steam_inlet.enth_mol[0]
        + m.fs.unit.feedwater_inlet.flow_mol[0] * +m.fs.unit.feedwater_inlet.enth_mol[0]
        - m.fs.unit.steam_outlet.flow_mol[0] * m.fs.unit.steam_outlet.enth_mol[0]
        - m.fs.unit.liquid_outlet.flow_mol[0] * m.fs.unit.liquid_outlet.enth_mol[0]
        + m.fs.unit.heat_duty[0]
    )
    # pressure drop
    assert pytest.approx(3662.5483, rel=1e-5) == pyo.value(m.fs.unit.deltaP[0])
    # mass balance
    assert pytest.approx(0, abs=1e-3) == pyo.value(
        m.fs.unit.water_steam_inlet.flow_mol[0]
        + m.fs.unit.feedwater_inlet.flow_mol[0]
        - m.fs.unit.steam_outlet.flow_mol[0]
        - m.fs.unit.liquid_outlet.flow_mol[0]
    )

    assert pytest.approx(
        pyo.value(
            m.fs.unit.water_steam_inlet.flow_mol[0]
            * m.fs.unit.flash.mixed_state[0].vapor_frac
        ),
        abs=1e-3,
    ) == pyo.value(m.fs.unit.steam_outlet.flow_mol[0])



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for CustomScalerBase.

Author: Andrew Lee
"""
import pytest
import re

from pyomo.environ import (
    Block,
    ComponentMap,
    ConcreteModel,
    Constraint,
    Set,
    Suffix,
    units,
    value,
    Var,
)
from pyomo.common.config import ConfigDict

from idaes.core.scaling.custom_scaler_base import (
    CustomScalerBase,
    ConstraintScalingScheme,
)
from idaes.core.util.constants import Constants
from idaes.core.util.testing import PhysicalParameterTestBlock
import idaes.logger as idaeslog


class DummyScaler(CustomScalerBase):
    """
    Create some dummy methods that will record that they ran in lists
    """

    def variable_scaling_routine(self, model, overwrite, submodel_scalers):
        # Create some lists to show that each method was run
        model._verification = []
        model._subscalers = {}
        model.overwrite = []

        model._verification.append("variables")
        model._subscalers["variables"] = submodel_scalers
        model.overwrite.append(overwrite)

    def constraint_scaling_routine(self, model, overwrite, submodel_scalers):
        model._verification.append("constraints")
        model._subscalers["constraints"] = submodel_scalers
        model.overwrite.append(overwrite)

    def fill_in_1(self, model):
        model._verification.append("fill_in_1")

    def fill_in_2(self, model):
        model._verification.append("fill_in_2")

    def dummy_method(self, model, overwrite):
        model._dummy_scaler_test = overwrite


@pytest.fixture
def model():
    m = ConcreteModel()

    m.pressure = Var(units=units.Pa, bounds=(1e4, None))
    m.temperature = Var(units=units.K, bounds=(250, 350))
    m.volume_mol = Var(units=units.m**3 / units.mol, bounds=(None, 10))
    m.enth_mol = Var(units=units.J / units.mol)

    m.ideal_gas = Constraint(
        expr=m.pressure * m.volume_mol == Constants.gas_constant * m.temperature
    )
    m.enthalpy_eq = Constraint(
        expr=4.81 * units.J / units.mol / units.K * m.temperature == m.enth_mol
    )

    m.scaling_factor = Suffix(direction=Suffix.EXPORT)

    return m


class TestCustomScalerBase:
    @pytest.mark.unit
    def test_init(self):
        sb = CustomScalerBase()

        assert isinstance(sb.config, ConfigDict)
        assert sb.config.zero_tolerance == 1e-12
        assert sb.config.max_variable_scaling_factor == 1e10
        assert sb.config.min_variable_scaling_factor == 1e-10
        assert sb.config.max_constraint_scaling_factor == 1e10
        assert sb.config.min_constraint_scaling_factor == 1e-10
        assert not sb.config.overwrite

        assert sb.default_scaling_factors == {}
        assert sb.unit_scaling_factors == {
            "Temperature": (units.K, 1e-2),
            "Pressure": (units.Pa, 1e-5),
        }

    @pytest.mark.unit
    def test_variable_scaling_routine(self):
        sb = CustomScalerBase()

        with pytest.raises(
            NotImplementedError,
            match="Custom Scaler has not implemented a variable_scaling_routine method.",
        ):
            sb.variable_scaling_routine("foo")

    @pytest.mark.unit
    def test_constraint_scaling_routine(self):
        sb = CustomScalerBase()

        with pytest.raises(
            NotImplementedError,
            match="Custom Scaler has not implemented a constraint_scaling_routine method.",
        ):
            sb.constraint_scaling_routine("foo")

    @pytest.mark.unit
    def test_scale_model(self):
        # Dummy object to hold testing data
        model = ConcreteModel()

        sb = DummyScaler()
        sb.scale_model(
            model,
            first_stage_fill_in=[sb.fill_in_1],
            second_stage_fill_in=[sb.fill_in_2],
            submodel_scalers={"foo", "bar"},
        )

        # Check order methods were run in
        assert model._verification == [
            "variables",
            "fill_in_1",
            "constraints",
            "fill_in_2",
        ]

        # Check that overwrite was passed on
        model.overwrite == [False, False]

        # Check that submodel scalers were passed on to methods
        assert model._subscalers == {
            "variables": {"foo", "bar"},
            "constraints": {"foo", "bar"},
        }

    @pytest.mark.unit
    def test_get_default_scaling_factor(self, model, caplog):
        caplog.set_level(idaeslog.DEBUG, logger="idaes")
        sb = CustomScalerBase()

        # No defaults defined yet
        assert sb.get_default_scaling_factor(model.pressure) is None
        assert "No default scaling factor found for pressure" in caplog.text

        # Set a default
        sb.default_scaling_factors["pressure"] = 1e-4
        assert sb.get_default_scaling_factor(model.pressure) == 1e-4

    @pytest.mark.unit
    def test_get_default_scaling_factor_indexed(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger="idaes")
        m = ConcreteModel()
        m.v = Var([1, 2, 3, 4])

        sb = CustomScalerBase()

        # No defaults defined yet
        assert sb.get_default_scaling_factor(m.v[1]) is None
        assert "No default scaling factor found for v[1]" in caplog.text

        # Set a default for the indexed var
        sb.default_scaling_factors["v"] = 1e-4
        assert sb.get_default_scaling_factor(m.v[1]) == 1e-4

        # Set a default for the specific element
        sb.default_scaling_factors["v[1]"] = 1e-8
        assert sb.get_default_scaling_factor(m.v[1]) == 1e-8

    @pytest.mark.unit
    def test_scale_variable_by_component(self, model, caplog):
        caplog.set_level(idaeslog.DEBUG, logger="idaes")
        sb = CustomScalerBase()

        # No scaling factors set
        sb.scale_variable_by_component(model.pressure, model.temperature)
        assert sb.get_scaling_factor(model.pressure) is None
        assert (
            "Could not set scaling factor for pressure, no scaling factor set for temperature"
            in caplog.text
        )

        # Set a scaling factor for temperature
        model.scaling_factor = Suffix(direction=Suffix.EXPORT)
        model.scaling_factor[model.temperature] = 1e-2

        sb.scale_variable_by_component(model.pressure, model.temperature)
        assert model.scaling_factor[model.pressure] == 1e-2

        # Change temperature scaling and check for overwrite
        model.scaling_factor[model.temperature] = 1e-3

        sb.scale_variable_by_component(
            model.pressure, model.temperature, overwrite=False
        )
        assert model.scaling_factor[model.pressure] == 1e-2

        sb.scale_variable_by_component(
            model.pressure, model.temperature, overwrite=True
        )
        assert model.scaling_factor[model.pressure] == 1e-3

    @pytest.mark.unit
    def test_scale_variable_by_bounds(self, model, caplog):
        caplog.set_level(idaeslog.DEBUG, logger="idaes")
        sb = CustomScalerBase()

        # Set a scaling factor for temperature to test overwrite
        model.scaling_factor[model.temperature] = 1e-2

        sb.scale_variable_by_bounds(model.temperature, overwrite=False)
        assert model.scaling_factor[model.temperature] == 1e-2

        # Both bounds
        sb.scale_variable_by_bounds(model.temperature, overwrite=True)
        assert model.scaling_factor[model.temperature] == 1 / 300

        # Lower bound only
        sb.scale_variable_by_bounds(model.pressure)
        assert model.scaling_factor[model.pressure] == 1e-4

        # Upper bound only
        sb.scale_variable_by_bounds(model.volume_mol)
        assert model.scaling_factor[model.volume_mol] == 0.1

        # No bounds
        sb.scale_variable_by_bounds(model.enth_mol)
        assert (
            "No scaling factor set for enth_mol; variable has no bounds." in caplog.text
        )
        assert model.enth_mol not in model.scaling_factor

        # Test case where bounds give a magnitude of 0
        model.pressure.setlb(-1e4)
        model.pressure.setub(1e4)
        sb.scale_variable_by_bounds(model.pressure, overwrite=True)
        assert model.scaling_factor[model.pressure] == 1

    @pytest.mark.unit
    def test_scale_variable_by_default(self, model, caplog):
        caplog.set_level(idaeslog.DEBUG, logger="idaes")
        sb = CustomScalerBase()

        # No defaults defined yet
        sb.scale_variable_by_default(model.pressure)
        assert model.pressure not in model.scaling_factor
        assert (
            "Could not set scaling factor for pressure, no default scaling factor set."
            in caplog.text
        )

        # Set a default
        sb.default_scaling_factors["pressure"] = 1e-4
        sb.scale_variable_by_default(model.pressure)
        assert model.scaling_factor[model.pressure] == 1e-4

        # Change default to check overwrite
        sb.default_scaling_factors["pressure"] = 1e-5
        sb.scale_variable_by_default(model.pressure, overwrite=False)
        assert model.scaling_factor[model.pressure] == 1e-4
        sb.scale_variable_by_default(model.pressure, overwrite=True)
        assert model.scaling_factor[model.pressure] == 1e-5

    @pytest.mark.unit
    def test_scale_variable_by_units(self, model, caplog):
        caplog.set_level(idaeslog.DEBUG, logger="idaes")
        sb = CustomScalerBase()

        # Units in dict, no conversion required
        sb.scale_variable_by_units(model.pressure)
        assert model.scaling_factor[model.pressure] == 1e-5

        # Equivalent units in dict, need conversion
        model.p2 = Var(units=units.bar)
        sb.scale_variable_by_units(model.p2)
        assert model.scaling_factor[model.p2] == pytest.approx(1, rel=1e-8)

        # No match - no sf assigned
        sb.scale_variable_by_units(model.enth_mol)
        assert model.enth_mol not in model.scaling_factor
        assert (
            "No scaling factor set for enth_mol; no match for units J/mol found "
            "in self.unit_scaling_factors" in caplog.text
        )

        # Test for overwrite
        model.scaling_factor[model.temperature] = 42
        sb.scale_variable_by_units(model.temperature, overwrite=False)
        assert model.scaling_factor[model.temperature] == 42
        sb.scale_variable_by_units(model.temperature, overwrite=True)
        assert model.scaling_factor[model.temperature] == 1e-2

    @pytest.mark.unit
    def test_scale_constraint_by_default(self, model, caplog):
        caplog.set_level(idaeslog.DEBUG, logger="idaes")
        sb = CustomScalerBase()

        # No defaults defined yet
        sb.scale_constraint_by_default(model.ideal_gas)
        assert model.ideal_gas not in model.scaling_factor
        assert (
            "Could not set scaling factor for ideal_gas, no default scaling factor set."
            in caplog.text
        )

        # Set a default
        sb.default_scaling_factors["ideal_gas"] = 1e-3
        sb.scale_constraint_by_default(model.ideal_gas)
        assert model.scaling_factor[model.ideal_gas] == 1e-3

        # Change default to check overwrite
        sb.default_scaling_factors["ideal_gas"] = 1e-5
        sb.scale_constraint_by_default(model.ideal_gas, overwrite=False)
        assert model.scaling_factor[model.ideal_gas] == 1e-3
        sb.scale_constraint_by_default(model.ideal_gas, overwrite=True)
        assert model.scaling_factor[model.ideal_gas] == 1e-5

    @pytest.mark.unit
    def test_get_expression_nominal_values(self, model):
        sb = CustomScalerBase()

        # Set variable scaling factors for testing
        model.scaling_factor[model.pressure] = 1e-5
        model.scaling_factor[model.temperature] = 1e-2
        model.scaling_factor[model.volume_mol] = 1e-1

        nominal_values = sb.get_expression_nominal_values(model.ideal_gas.expr)

        # Nominal values will be (R*T, P*V)
        assert nominal_values == [
            pytest.approx(831.446, rel=1e-5),
            pytest.approx(1e6, rel=1e-5),
        ]

        # Check redirection for ConstraintData objects
        nominal_values = sb.get_expression_nominal_values(model.ideal_gas)
        assert nominal_values == [
            pytest.approx(831.446, rel=1e-5),
            pytest.approx(1e6, rel=1e-5),
        ]

    @pytest.mark.unit
    def test_scale_constraint_by_nominal_value_harmonic(self, model):
        sb = CustomScalerBase()

        # Set variable scaling factors for testing
        model.scaling_factor[model.pressure] = 1e-5
        model.scaling_factor[model.temperature] = 1e-2
        model.scaling_factor[model.volume_mol] = 1e-1
        model.scaling_factor[model.ideal_gas] = 1

        # overwrite = False, no change
        sb.scale_constraint_by_nominal_value(
            model.ideal_gas,
            scheme=ConstraintScalingScheme.harmonicMean,
            overwrite=False,
        )
        assert model.scaling_factor[model.ideal_gas] == 1
        # overwrite = True
        sb.scale_constraint_by_nominal_value(
            model.ideal_gas, scheme=ConstraintScalingScheme.harmonicMean, overwrite=True
        )
        assert model.scaling_factor[model.ideal_gas] == pytest.approx(
            (1 / 831.446 + 1e-6), rel=1e-5
        )

    @pytest.mark.unit
    def test_scale_constraint_by_nominal_value_max(self, model):
        sb = CustomScalerBase()

        # Set variable scaling factors for testing
        model.scaling_factor[model.pressure] = 1e-5
        model.scaling_factor[model.temperature] = 1e-2
        model.scaling_factor[model.volume_mol] = 1e-1
        model.scaling_factor[model.ideal_gas] = 1

        # overwrite = False, no change
        sb.scale_constraint_by_nominal_value(model.ideal_gas, overwrite=False)
        assert model.scaling_factor[model.ideal_gas] == 1
        # overwrite = True
        sb.scale_constraint_by_nominal_value(model.ideal_gas, overwrite=True)
        assert model.scaling_factor[model.ideal_gas] == pytest.approx(1e-6, rel=1e-5)

    @pytest.mark.unit
    def test_scale_constraint_by_nominal_value_min(self, model):
        sb = CustomScalerBase()

        # Set variable scaling factors for testing
        model.scaling_factor[model.pressure] = 1e-5
        model.scaling_factor[model.temperature] = 1e-2
        model.scaling_factor[model.volume_mol] = 1e-1
        model.scaling_factor[model.ideal_gas] = 1

        # overwrite = False, no change
        sb.scale_constraint_by_nominal_value(
            model.ideal_gas,
            scheme=ConstraintScalingScheme.inverseMinimum,
            overwrite=False,
        )
        assert model.scaling_factor[model.ideal_gas] == 1
        # overwrite = True
        sb.scale_constraint_by_nominal_value(
            model.ideal_gas,
            scheme=ConstraintScalingScheme.inverseMinimum,
            overwrite=True,
        )
        assert model.scaling_factor[model.ideal_gas] == pytest.approx(
            1 / 831.446, rel=1e-5
        )

    @pytest.mark.unit
    def test_scale_constraint_by_nominal_value_inv_sum(self, model):
        sb = CustomScalerBase()

        # Set variable scaling factors for testing
        model.scaling_factor[model.pressure] = 1e-5
        model.scaling_factor[model.temperature] = 1e-2
        model.scaling_factor[model.volume_mol] = 1e-1
        model.scaling_factor[model.ideal_gas] = 1

        # overwrite = False, no change
        sb.scale_constraint_by_nominal_value(
            model.ideal_gas, scheme=ConstraintScalingScheme.inverseSum, overwrite=False
        )
        assert model.scaling_factor[model.ideal_gas] == 1
        # overwrite = True
        sb.scale_constraint_by_nominal_value(
            model.ideal_gas, scheme=ConstraintScalingScheme.inverseSum, overwrite=True
        )
        assert model.scaling_factor[model.ideal_gas] == pytest.approx(
            1 / (831.446 + 1e6), rel=1e-5
        )

    @pytest.mark.unit
    def test_scale_constraint_by_nominal_value_rss(self, model):
        sb = CustomScalerBase()

        # Set variable scaling factors for testing
        model.scaling_factor[model.pressure] = 1e-5
        model.scaling_factor[model.temperature] = 1e-2
        model.scaling_factor[model.volume_mol] = 1e-1
        model.scaling_factor[model.ideal_gas] = 1

        # overwrite = False, no change
        sb.scale_constraint_by_nominal_value(
            model.ideal_gas, scheme=ConstraintScalingScheme.inverseRSS, overwrite=False
        )
        assert model.scaling_factor[model.ideal_gas] == 1
        # overwrite = True
        sb.scale_constraint_by_nominal_value(
            model.ideal_gas, scheme=ConstraintScalingScheme.inverseRSS, overwrite=True
        )
        assert model.scaling_factor[model.ideal_gas] == pytest.approx(
            1 / (831.446**2 + 1e6**2) ** 0.5, rel=1e-5
        )

    @pytest.mark.unit
    def test_scale_constraint_by_nominal_value_invalid_scheme(self, model):
        sb = CustomScalerBase()

        with pytest.raises(
            ValueError,
            match=re.escape(
                "Invalid value for 'scheme' argument (foo) in "
                "scale_constraint_by_nominal_value."
            ),
        ):
            sb.scale_constraint_by_nominal_value(model.ideal_gas, scheme="foo")

    @pytest.mark.unit
    def test_scale_constraint_by_nominal_derivative_1norm(self, model):
        sb = CustomScalerBase()

        # Set variable scaling factors for testing
        model.scaling_factor[model.pressure] = 1e-5
        model.scaling_factor[model.temperature] = 1e-2
        model.scaling_factor[model.volume_mol] = 1e-1
        model.scaling_factor[model.ideal_gas] = 1

        # overwrite = False, no change
        sb.scale_constraint_by_nominal_derivative_norm(
            model.ideal_gas, norm=1, overwrite=False
        )
        assert model.scaling_factor[model.ideal_gas] == 1
        # overwrite = True
        sb.scale_constraint_by_nominal_derivative_norm(
            model.ideal_gas, norm=1, overwrite=True
        )
        assert model.scaling_factor[model.ideal_gas] == pytest.approx(
            4.99792e-7, rel=1e-5  # (1/(8.314+1e6+1e6)
        )

        # Check for clean up
        assert model.pressure.value is None
        assert model.temperature.value is None
        assert model.pressure.value is None

    @pytest.mark.unit
    def test_scale_constraint_by_nominal_derivative_2norm(self, model):
        sb = CustomScalerBase()

        # Set variable scaling factors for testing
        model.scaling_factor[model.pressure] = 1e-5
        model.scaling_factor[model.temperature] = 1e-2
        model.scaling_factor[model.volume_mol] = 1e-1
        model.scaling_factor[model.ideal_gas] = 1

        # overwrite = False, no change
        sb.scale_constraint_by_nominal_derivative_norm(model.ideal_gas, overwrite=False)
        assert model.scaling_factor[model.ideal_gas] == 1
        # overwrite = True
        sb.scale_constraint_by_nominal_derivative_norm(model.ideal_gas, overwrite=True)
        assert model.scaling_factor[model.ideal_gas] == pytest.approx(
            1 / (8.314**2 + 1e6**2 + 1e6**2) ** 0.5, rel=1e-5
        )

        # Check for clean up
        assert model.pressure.value is None
        assert model.temperature.value is None
        assert model.pressure.value is None

    @pytest.mark.unit
    def test_propagate_state_scaling(self):
        # Dummy up two state blocks
        m = ConcreteModel()

        m.properties = PhysicalParameterTestBlock()

        m.state1 = m.properties.build_state_block([1, 2, 3])
        m.state2 = m.properties.build_state_block([1, 2, 3])

        # Set scaling factors on state1
        for t, sd in m.state1.items():
            sd.scaling_factor = Suffix(direction=Suffix.EXPORT)
            sd.scaling_factor[sd.temperature] = 100 * t
            sd.scaling_factor[sd.pressure] = 1e5 * t

            count = 1
            for j in sd.flow_mol_phase_comp.values():
                sd.scaling_factor[j] = 10 * t * count
                count += 1

        sb = CustomScalerBase()
        sb.propagate_state_scaling(m.state2, m.state1)

        for t, sd in m.state2.items():
            assert sd.scaling_factor[sd.temperature] == 100 * t
            assert sd.scaling_factor[sd.pressure] == 1e5 * t

            count = 1
            for j in sd.flow_mol_phase_comp.values():
                assert sd.scaling_factor[j] == 10 * t * count
                count += 1

        # Test for overwrite
        for t, sd in m.state1.items():
            sd.scaling_factor[sd.temperature] = 200 * t
            sd.scaling_factor[sd.pressure] = 2e5 * t

            count = 1
            for j in sd.flow_mol_phase_comp.values():
                sd.scaling_factor[j] = 20 * t * count
                count += 1

        sb.propagate_state_scaling(m.state2, m.state1, overwrite=False)

        for t, sd in m.state2.items():
            assert sd.scaling_factor[sd.temperature] == 100 * t
            assert sd.scaling_factor[sd.pressure] == 1e5 * t

            count = 1
            for j in sd.flow_mol_phase_comp.values():
                assert sd.scaling_factor[j] == 10 * t * count
                count += 1

    @pytest.mark.unit
    def test_call_submodel_scaler_method_no_scaler(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger="idaes")

        # Dummy up a nested model
        m = ConcreteModel()
        m.b = Block([1, 2, 3])

        sb = CustomScalerBase()
        sb.call_submodel_scaler_method(m.b, method="dummy_method", overwrite=True)

        for bd in m.b.values():
            assert not hasattr(bd, "_dummy_scaler_test")

        assert "No default Scaler set for b. Cannot call dummy_method." in caplog.text

    @pytest.mark.unit
    def test_call_submodel_scaler_method_default_scaler(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger="idaes")

        # Dummy up a nested model
        m = ConcreteModel()
        m.b = Block([1, 2, 3])
        for bd in m.b.values():
            bd.default_scaler = DummyScaler

        sb = CustomScalerBase()
        sb.call_submodel_scaler_method(m.b, method="dummy_method", overwrite=True)

        for bd in m.b.values():
            assert bd._dummy_scaler_test

        assert "Using default Scaler for b." in caplog.text

    @pytest.mark.unit
    def test_call_submodel_scaler_method_user_scaler(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger="idaes")

        # Dummy up a nested model
        m = ConcreteModel()
        m.b = Block([1, 2, 3])

        scaler_map = ComponentMap()
        scaler_map[m.b] = DummyScaler()

        sb = CustomScalerBase()
        sb.call_submodel_scaler_method(
            m.b,
            method="dummy_method",
            submodel_scalers=scaler_map,
            overwrite=False,
        )

        for bd in m.b.values():
            assert not bd._dummy_scaler_test

        assert "Using user-defined Scaler for b." in caplog.text

    @pytest.mark.unit
    def test_call_submodel_scaler_method_user_scaler_class(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger="idaes")

        # Dummy up a nested model
        m = ConcreteModel()
        m.b = Block([1, 2, 3])

        scaler_map = ComponentMap()
        scaler_map[m.b] = DummyScaler()

        sb = CustomScalerBase()
        sb.call_submodel_scaler_method(
            m.b,
            method="dummy_method",
            submodel_scalers=scaler_map,
            overwrite=False,
        )

        for bd in m.b.values():
            assert not bd._dummy_scaler_test

        assert "Using user-defined Scaler for b." in caplog.text



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

import pytest
import pyomo.environ as pyo

from idaes.core import FlowsheetBlock
from idaes.models.properties import iapws95
from idaes.models_extra.power_generation.unit_models import FWH0DDynamic
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core.solvers import get_solver

prop_available = iapws95.iapws95_available()

# Set up solver
solver = get_solver()


@pytest.fixture(scope="module")
def build_unit():
    # Create a Concrete Model as the top level object
    m = pyo.ConcreteModel()
    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(
        dynamic=True,
        time_set=[0, 60],
        time_units=pyo.units.s,
        default_property_package=iapws95.Iapws95ParameterBlock(),
    )
    m.fs.properties = m.fs.config.default_property_package
    m.fs.unit = FWH0DDynamic(
        has_desuperheat=True,
        has_drain_cooling=True,
        has_drain_mixer=True,
        condense={
            "cold_side": {"has_pressure_change": True},
            "hot_side": {"has_pressure_change": True},
            "has_holdup": True,
        },
        desuperheat={"dynamic": False},
        cooling={"dynamic": False, "has_holdup": False},
        property_package=m.fs.properties,
    )
    m.discretizer = pyo.TransformationFactory("dae.finite_difference")
    m.discretizer.apply_to(m, nfe=2, wrt=m.fs.time, scheme="BACKWARD")
    m.fs.unit.set_initial_condition()
    m.fs.unit.desuperheat.hot_side_inlet.flow_mol.fix(100)
    m.fs.unit.desuperheat.hot_side_inlet.flow_mol.unfix()
    m.fs.unit.desuperheat.hot_side_inlet.pressure.fix(201325)
    m.fs.unit.desuperheat.hot_side_inlet.enth_mol.fix(60000)
    m.fs.unit.drain_mix.drain.flow_mol.fix(1)
    m.fs.unit.drain_mix.drain.pressure.fix(201325)
    m.fs.unit.drain_mix.drain.enth_mol.fix(20000)
    m.fs.unit.cooling.cold_side_inlet.flow_mol.fix(400)
    m.fs.unit.cooling.cold_side_inlet.pressure.fix(101325)
    m.fs.unit.cooling.cold_side_inlet.enth_mol.fix(3000)
    m.fs.unit.condense.area.fix(600)
    m.fs.unit.condense.overall_heat_transfer_coefficient.fix(1010)
    m.fs.unit.desuperheat.area.fix(85)
    m.fs.unit.desuperheat.overall_heat_transfer_coefficient.fix(145)
    m.fs.unit.cooling.area.fix(100)
    m.fs.unit.cooling.overall_heat_transfer_coefficient.fix(675)
    m.fs.unit.condense.cold_side.deltaP[:].fix(0)
    m.fs.unit.condense.level.fix(0.275)
    m.fs.unit.condense.heater_diameter.fix(1.4)
    m.fs.unit.condense.vol_frac_shell.fix(0.675)
    m.fs.unit.condense.cond_sect_length.fix(6.4)
    m.fs.unit.condense.cold_side.volume.fix(1.3)

    return m


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.unit
def test_basic_build(build_unit):
    """Make a model and make sure it doesn't throw exception"""
    m = build_unit
    assert degrees_of_freedom(m) == 0
    # Check unit config arguments
    assert len(m.fs.unit.config) == 10
    assert m.fs.unit.config.has_desuperheat
    assert m.fs.unit.config.has_drain_cooling
    assert m.fs.unit.config.has_drain_mixer
    assert m.fs.unit.config.condense
    assert m.fs.unit.config.property_package is m.fs.properties


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize_unit(build_unit):
    initialization_tester(build_unit, dof=0)


# TODO
# @pytest.mark.integration
# def test_units(build_unit):
#     # ToDo: dynamic models cannot assert units consistent
#     # assert_units_consistent(build_unit)


@pytest.mark.integration
@pytest.mark.skipif(not prop_available, reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_fwh_model(build_unit):
    m = build_unit
    # need to initialize because integration test do not run sequentially
    m.fs.unit.initialize()
    solver.options = {
        "tol": 1e-7,
        "linear_solver": "ma27",
        "max_iter": 50,
    }
    # initial drain flow rate
    drain_flow0 = m.fs.unit.cooling.hot_side_outlet.flow_mol[0].value
    assert pytest.approx(67.20607, abs=1e-2) == pyo.value(
        m.fs.unit.cooling.hot_side_outlet.flow_mol[0]
    )

    # change drain flow rate to increase water level
    for t in m.fs.time:
        if t >= 30:
            m.fs.unit.cooling.hot_side_outlet.flow_mol[t].fix(drain_flow0 * 0.95)
    m.fs.unit.condense.level.unfix()
    m.fs.unit.condense.level[0].fix()
    solver.solve(m, tee=True)
    assert degrees_of_freedom(m) == 0
    assert pytest.approx(0.2759, abs=1e-2) == pyo.value(m.fs.unit.condense.level[60])
